from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable
from itertools import cycle, islice
from typing import TYPE_CHECKING, Generic, TypeAlias, TypeVar

from pandas import MultiIndex

from xlviews.chart.style import COLORS, MARKER_DICT

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from pandas import DataFrame


T = TypeVar("T")


def get_columns_default(
    data: DataFrame,
    columns: str | list[str],
    default: dict[Hashable, T] | list[T] | None = None,
) -> tuple[list[str], dict[Hashable, T]]:
    if isinstance(columns, str):
        columns = [columns]

    if default is None:
        return columns, {}

    if isinstance(default, dict):
        return columns, default

    data = data[columns].drop_duplicates()
    values = [tuple(t) for t in data.itertuples(index=False)]
    default = dict(zip(values, cycle(default), strict=False))

    return columns, default


def get_index(
    data: DataFrame,
    default: Iterable[Hashable] | None = None,
) -> dict[tuple[Hashable, ...], int]:
    data = data.drop_duplicates()
    values = [tuple(t) for t in data.itertuples(index=False)]

    if default is None:
        return dict(zip(values, range(len(data)), strict=True))

    index = {}
    current_index = 0

    for default_value in default:
        if isinstance(default_value, tuple):
            value = default_value
        elif isinstance(default_value, list):
            value = tuple(default_value)
        else:
            value = (default_value,)

        if value in values:
            index[value] = current_index
            current_index += 1

    for value in values:
        if value not in index:
            index[value] = current_index
            current_index += 1

    return index


class Palette(ABC, Generic[T]):
    """A palette of items."""

    columns: list[str]
    index: dict[tuple[Hashable, ...], int]
    items: list[T]

    def __init__(
        self,
        data: DataFrame,
        columns: str | list[str],
        default: dict[Hashable, T] | list[T] | None = None,
    ) -> None:
        self.columns, default = get_columns_default(data, columns, default)
        self.index = get_index(data[self.columns], default)
        defaults = default.values()

        n = len(self.index) - len(default)
        self.items = [*defaults, *islice(self.cycle(defaults), n)]

    @abstractmethod
    def cycle(self, defaults: Iterable[T]) -> Iterator[T]:
        """Generate an infinite iterator of items."""

    def get(self, value: Hashable) -> int:
        if not isinstance(value, tuple):
            value = (value,)

        return self.index[value]

    def __getitem__(self, key: dict) -> T:
        if key == {None: 0}:  # from series
            return self.items[0]

        value = tuple(key[k] for k in self.columns)

        return self.items[self.get(value)]


class MarkerPalette(Palette[str]):
    def cycle(self, defaults: Iterable[str]) -> Iterator[str]:
        """Generate an infinite iterator of markers."""
        return cycle_markers(defaults)


def cycle_markers(skips: Iterable[str] | None = None) -> Iterator[str]:
    """Cycle through the markers."""
    if skips is None:
        skips = []

    markers = (m for m in MARKER_DICT if m != "")
    for marker in cycle(markers):
        if marker not in skips:
            yield marker


class ColorPalette(Palette[str]):
    def cycle(self, defaults: Iterable[str]) -> Iterator[str]:
        """Generate an infinite iterator of colors."""
        return cycle_colors(defaults)


def cycle_colors(skips: Iterable[str] | None = None) -> Iterator[str]:
    """Cycle through the colors."""
    if skips is None:
        skips = []

    for color in cycle(COLORS):
        if color not in skips:
            yield color


class FunctionPalette(Generic[T]):
    columns: str | list[str]
    func: Callable[[Hashable], T]

    def __init__(self, columns: str | list[str], func: Callable[[Hashable], T]) -> None:
        self.columns = columns
        self.func = func

    def __getitem__(self, key: dict) -> T:
        if isinstance(self.columns, str):
            return self.func(key[self.columns])

        value = tuple(key[k] for k in self.columns)
        return self.func(value)


PaletteStyle: TypeAlias = (
    str
    | list[str]
    | dict[Hashable, str]
    | Callable[[Hashable], str]
    | tuple[str | list[str], list[str] | dict[Hashable, str]]
    | tuple[str | list[str], Callable[[Hashable], str]]
    | Palette
    | FunctionPalette
)


def get_palette(
    cls: type[Palette],
    data: DataFrame,
    style: PaletteStyle | None,
) -> Palette | FunctionPalette | None:
    """Get a palette from a style."""
    if isinstance(style, Palette | FunctionPalette):
        return style

    if style is None:
        return None

    if isinstance(style, Callable):
        if isinstance(data.index, MultiIndex):
            return FunctionPalette(data.index.names, style)
        return FunctionPalette(data.index.name, style)

    if data.index.name is not None or isinstance(data.index, MultiIndex):
        data = data.index.to_frame(index=False)

    if isinstance(style, dict):
        return cls(data, data.columns.to_list(), style)

    if isinstance(style, tuple):
        columns, default = style
        if callable(default):
            return FunctionPalette(columns, default)

        return cls(data, columns, default)

    columns = style

    if isinstance(columns, str):
        columns = [columns]

    if any(c not in data for c in columns):
        data = data.drop_duplicates()
        values = [tuple(t) for t in data.itertuples(index=False)]
        default = dict(zip(values, cycle(columns), strict=False))
        return cls(data, data.columns.tolist(), default)  # type: ignore

    return cls(data, columns)


def get_marker_palette(
    data: DataFrame,
    marker: PaletteStyle | None,
) -> MarkerPalette | FunctionPalette | None:
    return get_palette(MarkerPalette, data, marker)  # type: ignore


def get_color_palette(
    data: DataFrame,
    color: PaletteStyle | None,
) -> ColorPalette | FunctionPalette | None:
    return get_palette(ColorPalette, data, color)  # type: ignore



import pytest
from pandas import DataFrame
from xlwings import Range, Sheet

from xlviews.core.formula import NONCONST_VALUE
from xlviews.dataframes.table import Table
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def df():
    return DataFrame({"a": [1, 1, 1, 0], "b": [2, 2, 3, 3], "c": [4, 4, 5, 6]})


@pytest.fixture(scope="module")
def rng(df: DataFrame, sheet_module: Sheet):
    rng = sheet_module.range("B3")
    rng.options(DataFrame, header=1, index=False).value = df
    return rng.expand()


def test_range_address(rng: Range):
    assert rng.get_address() == "$B$3:$D$7"


@pytest.fixture(scope="module")
def table(rng: Range):
    return Table(rng)


def test_cell(table: Table):
    assert table.cell.get_address() == "$B$3"


def test_column(table: Table):
    assert table.column.get_address() == "$B$4:$B$7"


def test_header(table: Table):
    assert table.header.get_address() == "$B$3:$D$3"


def test_const_header(table: Table):
    assert table.const_header.get_address() == "$B$2:$D$2"


def test_columns(table: Table):
    assert table.columns == ["a", "b", "c"]


def test_add_const_header(table: Table):
    table.add_const_header()
    assert table.const_header.value == [NONCONST_VALUE, NONCONST_VALUE, NONCONST_VALUE]


def test_add_const_header_clear(table: Table):
    table.add_const_header(clear=True)
    assert table.const_header.value == [None, None, None]


def test_from_api(table: Table):
    api = list(table.cell.sheet.api.ListObjects)[0]
    table2 = Table(api=api, sheet=table.sheet)
    assert table2.cell.row == table.cell.row
    assert table2.cell.column == table.cell.column


def test_unlist(df: DataFrame, sheet: Sheet):
    rng = sheet.range("B100")
    rng.options(DataFrame, header=1, index=False).value = df
    table = Table(rng.expand())
    assert len(list(sheet.api.ListObjects)) == 1
    table.unlist()
    assert not list(sheet.api.ListObjects)


@pytest.mark.parametrize(
    ("name", "value", "const"),
    [
        ("a", 1, [1, NONCONST_VALUE, NONCONST_VALUE]),
        ("a", 0, [0, 3, 6]),
        ("b", 2, [1, 2, 4]),
        ("b", 3, [NONCONST_VALUE, 3, NONCONST_VALUE]),
        ("c", 4, [1, 2, 4]),
        ("c", 5, [1, 3, 5]),
        ("c", 6, [0, 3, 6]),
        ("c", (4, 5), [1, NONCONST_VALUE, NONCONST_VALUE]),
        ("c", [5, 4], [1, NONCONST_VALUE, NONCONST_VALUE]),
        ("c", [4, 6], [NONCONST_VALUE, NONCONST_VALUE, NONCONST_VALUE]),
    ],
)
@pytest.mark.parametrize("to_dict", [True, False])
def test_auto_filter(table: Table, name, value, const, to_dict):
    table.add_const_header()

    if to_dict:
        table.auto_filter({name: value})
    else:
        table.auto_filter(name, value)

    assert table.const_header.value == const
    table.auto_filter(clear=True)
    table.add_const_header(clear=True)


def test_table_error():
    msg = "Either range or sheet and api must be provided"
    with pytest.raises(ValueError, match=msg):
        Table()



import pytest
from xlwings import Sheet
from xlwings.constants import ChartType, MarkerStyle

from xlviews.chart.axes import Axes
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available
from xlviews.testing.chart import Base

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize(
    ("label", "key", "expected"),
    [
        ("a", {}, "a"),
        ("a{b}", {"b": "B"}, "aB"),
        ("{a}{b}", {"a": "A", "b": "B"}, "AB"),
        (lambda x: f"_{x['a']}_", {"a": "A"}, "_A_"),
    ],
)
def test_get_label(label, key, expected):
    from xlviews.figure.plot import get_label

    assert get_label(label, key) == expected


def test_get_label_error():
    from xlviews.figure.plot import get_label

    with pytest.raises(ValueError, match="Invalid label"):
        get_label(1, 1)  # type: ignore


@pytest.fixture
def sf(sheet: Sheet):
    fc = Base(sheet, style=True)
    return fc.sf


@pytest.fixture
def ax(sf: SheetFrame):
    return Axes(2, 8, sheet=sf.sheet)


def test_plot_series(ax: Axes, sf: SheetFrame):
    from xlviews.figure.plot import Plot

    data = sf.agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatter)
        .set(label="abc", marker="o", color="blue", alpha=0.6)
    )
    s = p.series_collection[0]
    assert s.label == "abc"
    assert s.api.MarkerStyle == MarkerStyle.xlMarkerStyleCircle


def test_plot_index(ax: Axes, sf: SheetFrame):
    from xlviews.figure.plot import Plot

    data = sf.groupby("b").agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatterLines)
        .set(label="b={b}", marker=["o", "s"], color={"s": "red", "t": "blue"})
    )
    assert len(p.series_collection) == 2
    s = p.series_collection[0]
    assert s.label == "b=s"
    assert s.api.MarkerStyle == MarkerStyle.xlMarkerStyleCircle
    s = p.series_collection[1]
    assert s.label == "b=t"
    assert s.api.MarkerStyle == MarkerStyle.xlMarkerStyleSquare


def test_plot_multi_index(ax: Axes, sf: SheetFrame):
    from xlviews.figure.plot import Plot

    data = sf.groupby(["b", "c"]).agg(include_sheetname=True)
    p = (
        Plot(ax, data)
        .add("x", "y", ChartType.xlXYScatterLinesNoMarkers)
        .set(
            label=lambda x: f"{x['b']},{x['c']}",
            marker="b",
            color=("c", ["red", "green"]),
            size=10,
        )
    )
    assert len(p.series_collection) == 4
    s = p.series_collection[0]
    assert s.label == "s,100"
    assert s.api.MarkerStyle == MarkerStyle.xlMarkerStyleCircle
    s = p.series_collection[1]
    assert s.label == "s,200"
    assert s.api.MarkerStyle == MarkerStyle.xlMarkerStyleCircle
    s = p.series_collection[2]
    assert s.label == "t,100"
    assert s.api.MarkerStyle == MarkerStyle.xlMarkerStyleTriangle
    s = p.series_collection[3]
    assert s.label == "t,200"
    assert s.api.MarkerStyle == MarkerStyle.xlMarkerStyleTriangle



from __future__ import annotations

import numpy as np
import pandas as pd
from pandas import DataFrame

from xlviews.testing.common import FrameContainer, create_sheet


class NoIndex(FrameContainer):
    @classmethod
    def dataframe(cls) -> DataFrame:
        values = {"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}
        return DataFrame(values)


class Index(FrameContainer):
    row: int = 8
    column: int = 2

    @classmethod
    def dataframe(cls) -> DataFrame:
        values = {"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}
        index = ["x", "x", "y", "y"]
        df = DataFrame(values, index=index)
        df.index.name = "name"
        return df


class MultiIndex(FrameContainer):
    row: int = 2
    column: int = 6

    @classmethod
    def dataframe(cls) -> DataFrame:
        df = DataFrame(
            {
                "x": [1, 1, 1, 1, 2, 2, 2, 2],
                "y": [1, 1, 2, 2, 1, 1, 2, 2],
                "a": [1, 2, 3, 4, 5, 6, 7, 8],
                "b": [11, 12, 13, 14, 15, 16, 17, 18],
            },
        )
        return df.set_index(["x", "y"])


class MultiColumn(FrameContainer):
    row: int = 2
    column: int = 11

    @classmethod
    def dataframe(cls) -> DataFrame:
        a = ["a"] * 8 + ["b"] * 8
        b = (["c"] * 4 + ["d"] * 4) * 2
        c = np.repeat(range(1, 9), 2)
        d = ["x", "y"] * 8
        df = DataFrame(np.arange(16 * 6).reshape(16, 6).T)
        df.columns = pd.MultiIndex.from_arrays([a, b, c, d], names=["s", "t", "r", "i"])
        return df


class MultiIndexColumn(FrameContainer):
    row: int = 13
    column: int = 21

    @classmethod
    def dataframe(cls) -> DataFrame:
        df = DataFrame(
            {
                "x": [1, 1, 1, 1, 2, 2, 2, 2],
                "y": [1, 1, 2, 2, 1, 1, 2, 2],
                "z": [1, 1, 1, 2, 2, 1, 1, 1],
                "c": [1, 2, 3, 4, 5, 6, 7, 8],
                "d": [11, 12, 13, 14, 15, 16, 17, 18],
                "e": [21, 22, 23, 24, 25, 26, 27, 28],
                "f": [31, 32, 33, 34, 35, 36, 37, 38],
            },
        )
        df = df.set_index(["x", "y", "z"])
        x = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
        df.columns = pd.MultiIndex.from_tuples(x, names=["a", "b"])
        return df


class WideColumn(FrameContainer):
    row: int = 3
    column: int = 29

    @classmethod
    def dataframe(cls) -> DataFrame:
        x = ["i", "i", "j", "j", "i"]
        y = ["k", "l", "k", "l", "k"]
        a = list(range(5))
        b = list(range(10, 15))
        df = DataFrame({"x": x, "y": y, "a": a, "b": b})
        return df.set_index(["x", "y"])

    def init(self) -> None:
        self.sf.add_wide_column("u", list(range(3)))
        self.sf.add_wide_column("v", list(range(4)), style=True)


if __name__ == "__main__":
    sheet = create_sheet()

    classes: list[type[FrameContainer]] = [
        NoIndex,
        Index,
        MultiIndex,
        MultiColumn,
        MultiIndexColumn,
        WideColumn,
    ]

    fcs = [cls(sheet, style=True) for cls in classes]
    for fc in fcs:
        fc.sf.set_adjacent_column_width(1)

    fcs[3].sf.number_format("0.0", s="a", r=3)

    sf = fcs[-1].sf
    sf.autofit()

    fc = MultiIndex(sheet, row=15, style=True)
    sf = fc.sf
    sf.add_column("c", 10)
    sf.add_column("d", list(range(8)))
    sf.add_formula_column("e", "={c}+{d}")
    sf.style().autofit()
    sf.add_wide_column("u", list(range(3)))
    sf.add_wide_column("v", list(range(4)), number_format="0.0")
    sf.add_formula_column("u", "={u}+{a}")
    sf.add_formula_column("v", "={v}+{b}", style=True, autofit=True)
    sf.number_format(autofit=True, a="0.0", x="0.00")



import numpy as np
import pytest

from xlviews.colors import rgb
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.dataframes.stats_frame import StatsFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def funcs():
    return ["count", "mean", "median", "min", "max", "soa", "sum", "std"]


@pytest.fixture(scope="module")
def sf(sf_parent: SheetFrame, funcs: list[str]):
    return StatsFrame(sf_parent, funcs, by=":y")


def test_func(sf: StatsFrame, funcs: list[str]):
    rng = sf.sheet.range((sf.row + 1, sf.column), (sf.row + len(sf), sf.column))
    assert rng.value == np.tile(funcs, 4).tolist()


@pytest.mark.parametrize(
    ("func", "color"),
    [
        ("count", "gray"),
        ("mean", "#33aa33"),
        ("median", "#111111"),
        ("min", "#7777FF"),
        ("max", "#FF7777"),
        ("soa", "#5555FF"),
        ("sum", "purple"),
        ("std", "#aaaaaa"),
    ],
)
@pytest.mark.parametrize("c", ["func", "c"])
@pytest.mark.parametrize("o", [0, 8])
def test_color(sf: StatsFrame, funcs: list[str], func: str, color, c, o):
    row = funcs.index(func) + sf.row + o + 1
    column = sf.get_loc(c)
    rng = sf.sheet.range(row, column)
    assert rgb(rng.font.color) == rgb(color)


@pytest.mark.parametrize("func", ["soa", "sum"])
@pytest.mark.parametrize("c", ["func", "a"])
@pytest.mark.parametrize("o", [16, 24])
def test_italic(sf: StatsFrame, funcs: list[str], func: str, c, o):
    row = funcs.index(func) + sf.row + o + 1
    column = sf.get_loc(c)
    rng = sf.sheet.range(row, column)
    assert rng.font.italic


@pytest.mark.parametrize("c", ["a", "b"])
def test_soa(sf: StatsFrame, funcs: list[str], c):
    row = funcs.index("soa") + sf.row + 1
    column = sf.get_loc(c)
    rng = sf.sheet.range(row, column)
    assert rng.number_format == "0.0%"


@pytest.mark.parametrize("func", ["median", "min", "mean", "max", "std", "sum"])
@pytest.mark.parametrize("o", [0, 16])
def test_number_format(sf: StatsFrame, funcs: list[str], func, o):
    row = funcs.index(func) + sf.row + o + 1
    column = sf.get_loc("c")
    rng = sf.sheet.range(row, column)
    assert rng.number_format == "0.00"



from __future__ import annotations

from functools import cache


@cache
def index_to_column_name(n: int) -> str:
    """Return the Excel column name from an integer.

    Examples:
        >>> index_to_column_name(1)
        'A'
        >>> index_to_column_name(26)
        'Z'
        >>> index_to_column_name(27)
        'AA'
        >>> index_to_column_name(731)
        'ABC'
    """
    name = ""
    while n > 0:
        n -= 1
        name = chr(n % 26 + 65) + name
        n //= 26

    return name


@cache
def column_name_to_index(col: str) -> int:
    """Return the index from an Excel column name.

    Examples:
        >>> column_name_to_index("A")
        1
        >>> column_name_to_index("Z")
        26
        >>> column_name_to_index("AA")
        27
        >>> column_name_to_index("ABC")
        731
    """
    index = 0
    for char in col:
        index = index * 26 + (ord(char) - ord("A") + 1)

    return index



import numpy as np
import pytest
from pandas import DataFrame
from xlwings import Range as RangeImpl
from xlwings import Sheet

from xlviews.core.formula import NONCONST_VALUE
from xlviews.core.range import Range
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def df():
    return DataFrame({"a": [1, 1, 1, 1], "b": [2, 2, 3, 3], "c": [4, 4, 4, 4]})


@pytest.fixture(scope="module", params=[RangeImpl, Range], ids=["RangeImpl", "Range"])
def rng(df: DataFrame, sheet_module: Sheet, request: pytest.FixtureRequest):
    rng = sheet_module.range("B3")
    rng.options(DataFrame, header=1, index=False).value = df
    rng: RangeImpl = rng.expand()
    return request.param(
        (rng.row, rng.column),
        (rng.last_cell.row, rng.last_cell.column),
    )


def test_range_address(rng: Range | RangeImpl):
    assert rng.get_address() == "$B$3:$D$7"


@pytest.fixture(scope="module")
def column(rng: Range | RangeImpl):
    start = rng[0].offset(1)
    end = start.offset(3)
    return rng.__class__((start.row, start.column), (end.row, end.column))


def test_column(column: Range | RangeImpl):
    assert column.get_address() == "$B$4:$B$7"


@pytest.fixture(scope="module")
def const_header(rng: Range | RangeImpl):
    start = (rng[0].row, rng[0].column)
    rng_impl = rng if isinstance(rng, RangeImpl) else rng.impl
    end = rng_impl[0].expand("right")
    return rng.__class__(start, (end[-1].row, end[-1].column)).offset(-1)


def test_header(const_header: Range | RangeImpl):
    assert const_header.get_address() == "$B$2:$D$2"


@pytest.mark.parametrize(("k", "value"), [(0, 1), (1, NONCONST_VALUE), (2, 4)])
def test_const(column: Range | RangeImpl, const_header: Range | RangeImpl, k, value):
    from xlviews.core.formula import const

    const_header.value = "=" + const(column)
    assert const_header.value[k] == value


@pytest.fixture(scope="module")
def ranges(sheet_module: Sheet):
    rng = sheet_module.range("B100")
    rng.options(transpose=True).value = [1, 2, 3, 4, 0, 5, 6, 7, 8, 9, 10]
    rng1 = rng.expand("down")
    sheet_module.range("B104").value = None

    rng = sheet_module.range("C100")
    rng.options(transpose=True).value = [11, 12, 13, 14, 15, 16, 17, 0, 18, 19, 20]
    rng2 = rng.expand("down")
    sheet_module.range("C107").value = None

    return [Range.from_range(rng1), Range.from_range(rng2)]


def test_ranges(ranges: list[Range]):
    assert ranges[0].get_address() == "$B$100:$B$110"
    assert ranges[1].get_address() == "$C$100:$C$110"


def test_aggregate_value(ranges: list[Range]):
    from xlviews.core.formula import aggregate

    x = aggregate("count", ranges)
    assert x == "AGGREGATE(2,7,$B$100:$B$110,$C$100:$C$110)"


def test_aggregate_none(ranges: list[Range]):
    from xlviews.core.formula import aggregate

    x = aggregate(None, ranges)
    assert x == "$B$100:$B$110,$C$100:$C$110"


def test_aggregate_formula(ranges: list[Range]):
    from xlviews.core.formula import aggregate

    x = aggregate("max", ranges, formula=True)
    assert x == "=AGGREGATE(4,7,$B$100:$B$110,$C$100:$C$110)"


def test_aggreate_range_range(ranges: list[Range]):
    from xlviews.core.formula import aggregate

    assert aggregate("max", ranges[0]) == "AGGREGATE(4,7,$B$100:$B$110)"


FUNC_VALUES = [
    ("count", 20),
    ("sum", 210),
    ("min", 1),
    ("max", 20),
    ("mean", 10.5),
    ("median", 10.5),
    ("std", np.std(range(1, 21))),
    ("soa", np.std(range(1, 21)) / 10.5),
]


@pytest.mark.parametrize(("func", "value"), FUNC_VALUES)
def test_aggregate_str(ranges: list[Range], func, value):
    from xlviews.core.formula import aggregate

    cell = ranges[0].sheet.range("D100")
    cell.value = aggregate(func, ranges, formula=True)
    assert cell.value == value


@pytest.mark.parametrize(("func", "value"), FUNC_VALUES)
@pytest.mark.parametrize("apply", [lambda x: x, Range.from_range])
def test_aggregate_range(ranges: list[Range], func, value, apply):
    from xlviews.core.formula import aggregate

    ref = apply(ranges[0].sheet.range("E100"))
    ref.value = func
    formula = aggregate(ref, ranges)
    cell = ranges[0].sheet.range("D100")
    cell.value = "=" + formula
    assert cell.value == value


def test_aggreate_func_invalid():
    from xlviews.core.formula import aggregate

    with pytest.raises(ValueError, match="Invalid aggregate function: sin"):
        aggregate("sin", "A1:A10")



from __future__ import annotations

from typing import TYPE_CHECKING

from pandas import DataFrame, Index, MultiIndex

from xlviews.core.formula import aggregate
from xlviews.core.range import Range
from xlviews.dataframes.colorbar import Colorbar
from xlviews.style import set_color_scale, set_font
from xlviews.utils import suspend_screen_updates

from .sheet_frame import SheetFrame
from .style import set_heat_frame_style

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterator, Sequence
    from typing import Any, Literal, Self

    from pandas import Index
    from xlwings import Sheet

    from xlviews.colors import Color


class HeatFrame(SheetFrame):
    index: Index
    columns: Index
    range: Range

    @suspend_screen_updates
    def __init__(
        self,
        row: int,
        column: int,
        data: DataFrame,
        sheet: Sheet | None = None,
        vmin: float | str | Range | None = None,
        vmax: float | str | Range | None = None,
    ) -> None:
        data = clean_data(data)

        super().__init__(row, column, data, sheet)

        self.columns = data.columns

        start = self.row + 1, self.column + 1
        end = start[0] + self.shape[0] - 1, start[1] + self.shape[1] - 1
        self.range = Range(start, end, self.sheet)

        set_heat_frame_style(self)
        self.set(vmin, vmax)

    def set(
        self,
        vmin: float | str | Range | None = None,
        vmax: float | str | Range | None = None,
    ) -> Self:
        rng = self.range

        if vmin is None:
            vmin = aggregate("min", rng)
        if vmax is None:
            vmax = aggregate("max", rng)

        set_color_scale(rng, vmin, vmax)
        return self

    def style(
        self,
        *,
        size: float | None = None,
        bold: bool | None = None,
        italic: bool | None = None,
        color: Color | None = None,
    ) -> Self:
        set_font(self.range, size=size, bold=bold, italic=italic, color=color)
        return self

    def number_format(self, number_format: str) -> Self:
        self.range.impl.number_format = number_format
        return self

    def colorbar(
        self,
        vmin: float | str | Range | None = None,
        vmax: float | str | Range | None = None,
        label: str | None = None,
        autofit: bool = False,
    ) -> Colorbar:
        row = self.row + 1
        column = self.column + self.shape[1] + 2
        length = self.shape[0]

        if vmin is None:
            vmin = self.range
        if vmax is None:
            vmax = self.range

        cb = Colorbar(row, column, length, sheet=self.sheet)
        cb.set(vmin, vmax, label, autofit)
        return cb

    @classmethod
    def facet(
        cls,
        row: int,
        column: int,
        data: DataFrame,
        index: str | list[str] | None = None,
        columns: str | list[str] | None = None,
        padding: tuple[int, int] = (2, 1),
    ) -> Iterator[tuple[dict[Hashable, Any], Self]]:
        for r, ikey in iterrows(data.index, index, row, padding[0] + 1):
            for c, ckey in iterrows(data.columns, columns, column, padding[1] + 1):
                sub = xs(data, ikey, ckey)
                yield (ikey | ckey), cls(r, c, sub)

    @classmethod
    def pair(
        cls,
        row: int,
        column: int,
        data: DataFrame,
        values: str | list[str] | None = None,
        index: str | list[str] | None = None,
        columns: str | list[str] | None = None,
        padding: tuple[int, int] = (2, 1),
        value_name: str = "value",
        axis: Literal[0, 1] | None = None,
    ) -> Iterator[tuple[dict[Hashable, Any], Self]]:
        if values is None:
            values = data.columns.get_level_values(0).unique().to_list()
        elif isinstance(values, str):
            values = [values]

        if axis is None:
            axis = 1 if columns is None else 0

        nr = row
        nc = column

        for value in values:
            sub = xs(data, None, {0: value})

            for key, frame in cls.facet(row, column, sub, index, columns, padding):
                yield {value_name: value} | key, frame
                if axis == 1:
                    nc = max(nc, frame.column + frame.shape[1] + 1 + padding[1])
                else:
                    nr = max(nr, frame.row + frame.shape[0] + 1 + padding[0])

            column = nc
            row = nr


def clean_data(data: DataFrame) -> DataFrame:
    data = data.copy()

    if isinstance(data.columns, MultiIndex):
        data.columns = data.columns.droplevel(list(range(1, data.columns.nlevels)))

    if isinstance(data.index, MultiIndex):
        data.index = data.index.droplevel(list(range(1, data.index.nlevels)))

    data.index.name = None

    return data


def iterrows(
    index: Index,
    levels: int | str | Sequence[int | str] | None,
    offset: int = 0,
    padding: int = 0,
) -> Iterator[tuple[int, dict[Hashable, Any]]]:
    if levels is None:
        yield offset, {}
        return

    if isinstance(levels, int | str):
        levels = [levels]

    if levels:
        values = {level: index.get_level_values(level) for level in levels}
        it = DataFrame(values).drop_duplicates().iterrows()

        for k, (i, s) in enumerate(it):
            if not isinstance(i, int):
                raise NotImplementedError

            yield i + offset + k * padding, s.to_dict()


def xs(
    df: DataFrame,
    index: dict[Hashable, Any] | None,
    columns: dict[Hashable, Any] | None,
) -> DataFrame:
    if index:
        df = df.xs(tuple(index.values()), 0, tuple(index.keys()))  # type: ignore

    if columns:
        df = df.xs(tuple(columns.values()), 1, tuple(columns.keys()))  # type: ignore

    return df



import pytest
from xlwings import Sheet

from xlviews.core.range import Range
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def rng(sheet_module: Sheet):
    rng = Range((2, 2), (3, 3), sheet_module)
    rng.value = [[1, 2], [3, 4]]
    return rng


def test_colorbar_vertical(rng: Range, sheet_module: Sheet):
    from xlviews.dataframes.colorbar import Colorbar

    cb = Colorbar(2, 5, 6, sheet=sheet_module)
    cb.set(vmin=rng, vmax=rng, label="T", autofit=True)
    cb.set_adjacent_column_width(1)
    cb.apply(rng)

    assert sheet_module.range((1, 5)).value == "T"
    assert sheet_module.range((2, 5)).value == 4
    assert sheet_module.range((7, 5)).value == 1
    assert sheet_module.range((1, 6)).column_width == 1


def test_colorbar_horizontal(rng: Range, sheet_module: Sheet):
    from xlviews.dataframes.colorbar import Colorbar

    cb = Colorbar(2, 7, 10, orientation="horizontal", sheet=sheet_module)
    cb.set(vmin=rng, vmax=rng, label="T", autofit=True)
    cb.set_adjacent_column_width(1)

    assert sheet_module.range((2, 17)).value == "T"
    assert sheet_module.range((2, 16)).value == 4
    assert sheet_module.range((2, 7)).value == 1
    assert sheet_module.range((1, 18)).column_width == 1
